# coding: utf-8
# Copyright (c) Qotto, 2018

from .base import BaseEvent

from typing import Any, Dict

__all__ = [
    'GenericEvent'
]


class GenericEvent(BaseEvent):
    """
    Allows to create generic named events.

    >>> evt = GenericEvent(name='qotto.payment.command.CreatePayment', data={'quantity': '1L'})
    >>> evt.name
    'qotto.payment.command.CreatePayment'
    >>> evt.data['quantity']
    '1L'
    """

    def __init__(self, name: str, data: Dict[str, Any]) -> None:
        super().__init__(name=name, data=data)

    @classmethod
    def from_data(cls, event_name: str, event_data: Dict[str, Any]):
        event = cls(name=event_name, data=event_data)
        return event



# coding: utf-8
# Copyright (c) Qotto, 2018

import json
import os
import yaml

import logging
from ..event.base import BaseEvent
from ..event.generic import GenericEvent
from .base import BaseEventSerializer
from avro.datafile import DataFileWriter, DataFileReader
from avro.io import DatumWriter, DatumReader
from avro.schema import NamedSchema, Parse
from io import BytesIO
import re

from typing import Dict, Type

__all__ = [
    'AvroEventSerializer',
]


class AvroEventSerializer(BaseEventSerializer):
    AVRO_SCHEMA_FILE_EXTENSION = 'avsc.yaml'

    def __init__(self, settings: object) -> None:
        self.logger = logging.getLogger(__name__)
        self._schemas: Dict[str, NamedSchema] = dict()
        self._events: Dict[object, Type[BaseEvent]] = dict()

        if not hasattr(settings, 'AVRO_SCHEMAS_FOLDER'):
            raise Exception('Missing AVRO_SCHEMAS_FOLDER config')
        self.scan_folder(schemas_folder=settings.AVRO_SCHEMAS_FOLDER)

    def scan_folder(self, schemas_folder: str) -> None:
        """
        Searches Avro schema files in the specified folder.
        The schema files should have the avsc.yaml extension.
        """
        with os.scandir(schemas_folder) as entries:  # type: ignore
            for entry in entries:
                # ignores everything that is not a file
                if not entry.is_file():
                    continue
                # ignores hidden files
                if entry.name.startswith('.'):
                    continue
                # ignores everything files with wrong extension
                if not entry.name.endswith(f'.{self.AVRO_SCHEMA_FILE_EXTENSION}'):
                    continue
                self._extract_schema_from_file(entry.path)

    def _extract_schema_from_file(self, file_path: str) -> None:
        with open(file_path, 'r') as f:
            for s in yaml.load_all(f):
                avro_schema_data = json.dumps(s)
                avro_schema = Parse(avro_schema_data)
                schema_name = avro_schema.namespace + '.' + avro_schema.name
                if schema_name in self._schemas:
                    raise Exception(
                        f"Avro schema {schema_name} was defined more than once!")
                self._schemas[schema_name] = avro_schema

    def register_event_class(self, event_class: Type[BaseEvent], event_name: str) -> None:
        """
        Registers an event class associated to a particular event name.
        """

        # name can be a regex
        event_name_regex = re.compile(event_name)

        match = False
        for schema_name in self._schemas:
            if event_name_regex.match(schema_name):
                match = True
                break
        if not match:
            raise NameError(f"{event_name} does not match any schema")

        self._events[event_name_regex] = event_class

    def encode(self, event: BaseEvent) -> bytes:
        schema = self._schemas[event.name]

        if schema is None:
            raise NameError(
                f"No schema found to encode event with name {event.name}")

        output = BytesIO()
        writer = DataFileWriter(output, DatumWriter(), schema)
        writer.append(event.data)
        writer.flush()
        encoded_event = output.getvalue()
        writer.close()
        return encoded_event

    def decode(self, encoded_event: bytes) -> BaseEvent:
        reader = DataFileReader(BytesIO(encoded_event), DatumReader())
        schema = json.loads(reader.meta.get('avro.schema').decode('utf-8'))
        schema_name = schema['namespace'] + '.' + schema['name']
        event_data = next(reader)

        # finds a matching event name
        event_class = None
        for e_name, e_class in self._events.items():
            if e_name.match(schema_name):
                event_class = e_class
                break

        if event_class is None:
            event_class = GenericEvent

        return event_class.from_data(event_name=schema_name, event_data=event_data)



from aiokafka import AIOKafkaConsumer, TopicPartition
import logging
import sys

from typing import Any, Dict, Type, Callable, List
import asyncio
from ..serializer.base import BaseEventSerializer
from ..event.base import BaseEvent
from ..command.base import BaseCommand
from .base import BaseEventConsumer
from ..app.base import BaseApp

__all__ = [
    'KafkaConsumer'
]


class KafkaConsumer(BaseEventConsumer):

    def __init__(self, settings: object, app: BaseApp, serializer: BaseEventSerializer, event_topics: List[str], event_group: str, position: str) -> None:
        self.logger = logging.getLogger(__name__)

        if not hasattr(settings, 'KAFKA_BOOTSTRAP_SERVER'):
            raise Exception('Missing KAFKA_BOOTSTRAP_SERVER config')

        self.max_retries = 10
        if hasattr(settings, 'EVENTY_CONSUMER_MAX_RETRIES'):
            self.max_retries = settings.EVENTY_CONSUMER_MAX_RETRIES

        self.retry_interval = 1000
        if hasattr(settings, 'EVENTY_CONSUMER_RETRY_INTERVAL'):
            self.retry_interval = settings.EVENTY_CONSUMER_RETRY_INTERVAL

        self.retry_backoff_coeff = 2
        if hasattr(settings, 'EVENTY_CONSUMER_RETRY_BACKOFF_COEFF'):
            self.retry_backoff_coeff = settings.EVENTY_CONSUMER_RETRY_BACKOFF_COEFF

        self.app = app
        self.event_topics = event_topics
        self.event_group = event_group
        self.position = position
        self.consumer = None
        self.current_position_checkpoint_callback = None
        self.end_position_checkpoint_callback = None
        bootstrap_servers = settings.KAFKA_BOOTSTRAP_SERVER

        consumer_args: Dict[str, Any]
        consumer_args = {
            'loop': asyncio.get_event_loop(),
            'bootstrap_servers': [bootstrap_servers],
            'enable_auto_commit': False,
            'group_id': self.event_group,
            'value_deserializer': serializer.decode,
            'auto_offset_reset': self.position
        }

        try:
            self.consumer = AIOKafkaConsumer(
                *self.event_topics, **consumer_args)

        except Exception as e:
            self.logger.error(
                f"Unable to connect to the Kafka broker {bootstrap_servers} : {e}")
            raise e

    def set_current_position_checkpoint_callback(self, checkpoint_callback):
        self.current_position_checkpoint_callback = checkpoint_callback

    def set_end_position_checkpoint_callback(self, checkpoint_callback):
        self.end_position_checkpoint_callback = checkpoint_callback

    async def current_position(self):
        position = {}
        for partition in self.consumer.assignment():
            offset = await self.consumer.committed(partition) or 0
            position[partition] = offset

        return position

    async def end_position(self):
        position = {}
        for partition in self.consumer.assignment():
            offset = (await self.consumer.end_offsets([partition]))[partition]
            position[partition] = offset

        return position

    async def is_checkpoint_reached(self, checkpoint):
        for partition in self.consumer.assignment():
            position = (await self.consumer.position(partition))
            if position < checkpoint[partition]:
                return False
        return True

    async def start(self):

        self.logger.info(
            f'Starting kafka consumer on topic {self.event_topics} with group {self.event_group}')
        try:
            await self.consumer.start()
        except Exception as e:
            self.logger.error(
                f'An error occurred while starting kafka consumer on topic {self.event_topics} with group {self.event_group}: {e}')
            sys.exit(1)

        current_position_checkpoint = None
        end_position_checkpoint = None
        if self.event_group is not None:
            current_position = await self.current_position()
            end_position = await self.end_position()
            self.logger.debug(f'Current position : {current_position}')
            self.logger.debug(f'End position : {end_position}')

            if self.position == 'earliest' and self.event_group is not None:
                current_position_checkpoint = current_position
                end_position_checkpoint = end_position
                await self.consumer.seek_to_beginning()

        async for msg in self.consumer:

            retries = 0
            sleep_duration_in_ms = self.retry_interval
            while True:
                try:
                    event = msg.value
                    corr_id = event.correlation_id

                    self.logger.info(
                        f"[CID:{corr_id}] Start handling {event.name}")
                    await event.handle(app=self.app, corr_id=corr_id)
                    self.logger.info(
                        f"[CID:{corr_id}] End handling {event.name}")

                    if self.event_group is not None:
                        self.logger.debug(
                            f"[CID:{corr_id}] Commit Kafka transaction")
                        await self.consumer.commit()

                    self.logger.debug(
                        f"[CID:{corr_id}] Continue with the next message")
                    # break the retry loop
                    break

                except Exception as e:
                    self.logger.exception(
                        f'[CID:{corr_id}] An error occurred while handling received message.')

                    if retries != self.max_retries:
                        # increase the number of retries
                        retries = retries + 1

                        sleep_duration_in_s = int(sleep_duration_in_ms/1000)
                        self.logger.info(
                            f"[CID:{corr_id}] Sleeping {sleep_duration_in_s}s a before retrying...")
                        await asyncio.sleep(sleep_duration_in_s)

                        # increase the sleep duration
                        sleep_duration_in_ms = sleep_duration_in_ms * self.retry_backoff_coeff

                    else:
                        self.logger.error(
                            f'[CID:{corr_id}] Unable to handle message within {1 + self.max_retries} tries. Stopping process')
                        sys.exit(1)

            if current_position_checkpoint and await self.is_checkpoint_reached(current_position_checkpoint):
                self.logger.info('Current position checkpoint reached')
                if self.current_position_checkpoint_callback:
                    await self.current_position_checkpoint_callback()
                current_position_checkpoint = None

            if end_position_checkpoint and await self.is_checkpoint_reached(end_position_checkpoint):
                self.logger.info('End position checkpoint reached')
                if self.end_position_checkpoint_callback:
                    await self.end_position_checkpoint_callback()
                end_position_checkpoint = None



from typing import Type, Callable
from ..event.base import BaseEvent

__all__ = [
    'BaseEventConsumer'
]


class BaseEventConsumer:
    async def start(self):
        raise NotImplementedError

    def set_checkpoint_callback(self, checkpoint_callback):
        raise NotImplementedError
